/*
This program is confidential and proprietary to MRJ Technology Solutions
and may not be reproduced, published or disclosed to others without
written authorization from MRJ.

Copyright (c) 1998 MRJ Technology Solutions, Inc. All Rights Reserved.

This program is derived from prior work some of which was originally
written as a joint project between the Numerical Aerospace Simulation
(NAS) Systems Division of NASA Ames Research Center and the National
Energy Research Supercomputer Center (NERSC) of Lawrence Livermore
National Laboratory.
*/
/*
 * queue_recov.c - This file contains the functions to record a queue
 *	data struture to disk and to recover it from disk.
 *
 *	The data is recorded in a file whose name is the queue name
 *
 *	The following public functions are provided:
 *		que_save()  - save the disk image 
 *		que_recov()  - recover (read) queue from disk
 */

#include <pbs_config.h>   /* the master config generated by configure */

#include <sys/types.h>
#include <sys/param.h>
#include "pbs_ifl.h"
#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include "server_limits.h"
#include "list_link.h"
#include "attribute.h"
#include "queue.h"
#include "svrfunc.h"

static char ident[] = "@(#) $RCSfile: queue_recov.c,v $ $Revision: 2.0 $";

/*
 * the following funny business is due to the fact that O_SYNC 
 * is not currently POSIX
 */
#ifdef O_SYNC
#define O_Sync O_SYNC
#elif _FSYNC
#define O_Sync _FSYNC
#else
#define O_Sync 0
#endif

/* data global to this file */

extern char *path_queues;

/*
 * que_save() - Saves a queue structure image on disk
 *
 *
 *	For a save, to insure no data is ever lost due to system crash:
 *	1. write (with O_SYNC) new image to a new file using a temp name
 *	2. unlink the old (image) file
 *	3. link the correct name to the new file
 *	4. unlink the temp name
 *	
 *	Then, if the queue has any access control lists, they are saved
 *	to their own files.
 */

int que_save (pque)
	pbs_queue	*pque;		/* pointer to queue structure */
{
	int	fds;
	int	i;
	char   *myid = "que_save";
	char	namebuf1[MAXPATHLEN];
	char	namebuf2[MAXPATHLEN];

	(void)strcpy(namebuf1, path_queues);
	(void)strcat(namebuf1, pque->qu_qs.qu_name);
	(void)strcpy(namebuf2, namebuf1);
	(void)strcat(namebuf2, ".new");
	fds = open(namebuf2, O_CREAT | O_WRONLY | O_Sync, 0600);
	if (fds < 0) {
		log_err(errno, myid, "open error");
		return (-1);
	}

	/* set up save buffering system */

	save_setup(fds);

	/* save basic queue structure (fixed length stuff) */

	if (save_struct((char *)&pque->qu_qs, sizeof (struct queuefix)) != 0) {
		log_err(-1, myid, "save_struct failed");
		(void)close(fds);
		return (-1);
	}

	/* save queue attributes  */

	if (save_attr(que_attr_def,pque->qu_attr,(int)QA_ATR_LAST) != 0) {
		log_err(-1, myid, "save_attr failed");
		(void)close(fds);
		return (-1);
	}

	if (save_flush() != 0) {	/* flush the save buffer */
		log_err(-1, myid,"save_flush failed");
		(void)close(fds);
		return (-1);
	}

	(void)close(fds);

	(void)unlink(namebuf1);	
	if (link(namebuf2, namebuf1) < 0) {
		log_err(errno, myid, "unable to link queue name");
	} else {
		(void)unlink(namebuf2);
	}
	
	/* 
	 * now search queue's attributes for access control lists,
	 * they are saved separately in their own file:
	 * ../priv/(attr name)/(queue name)
	 */

	for (i=0; i < (int)QA_ATR_LAST; i++) {
		if (pque->qu_attr[i].at_type == ATR_TYPE_ACL) {
			save_acl(&pque->qu_attr[i], &que_attr_def[i],
				 que_attr_def[i].at_name, pque->qu_qs.qu_name);
		}
	}
	
	return (0);
}

/*
 * que_recov() - load (recover) a queue from its save file
 *
 *	This function is only needed upon server start up.
 *
 *	The queue structure is recovered from the disk.
 *	Space to hold the above is malloc-ed as needed.
 *
 *	Returns: pointer to new queue structure if successful
 *		 null if error
 */

pbs_queue *que_recov(filename)
	char   *filename;	/* pathname to queue save file */
{
	int	   fds;
	int	   i;
	pbs_queue *pq;
	char	   namebuf[MAXPATHLEN];


	pq = que_alloc(filename);  /* allocate & init queue structure space */
	if (pq == (pbs_queue *)0) {
		log_err(-1, "que_recov", "que_alloc failed");
		return ((pbs_queue *)0);
	}

	(void)strcpy(namebuf, path_queues);
	(void)strcat(namebuf, filename);
	fds = open(namebuf, O_RDONLY, 0);
	if (fds < 0) {
		log_err(errno, "que_recov", "open error");
		free((char *)pq);
		return ((pbs_queue *)0);
	}

	/* read in queue save sub-structure */

	if (read(fds, (char *)&pq->qu_qs, sizeof(struct queuefix)) !=
	    sizeof(struct queuefix)) {
		log_err(errno, "que_recov", "read error");
		free((char *)pq);
		(void)close(fds);
		return ((pbs_queue *)0);
	}

	/* read in queue attributes */

	if (recov_attr(fds, pq, que_attr_def, pq->qu_attr,
	    (int)QA_ATR_LAST, 0) != 0) {
		log_err(-1, "que_recov", "recov_attr[common] failed");
		que_free(pq);
		(void)close(fds);
		return ((pbs_queue *)0);
	}

	/*
	 * now reload the access control lists, these attributes were
	 * saved separately
	 */

	for (i=0; i < (int)QA_ATR_LAST; i++) {
		if (pq->qu_attr[i].at_type == ATR_TYPE_ACL) {
			recov_acl(&pq->qu_attr[i], &que_attr_def[i],
			         que_attr_def[i].at_name, pq->qu_qs.qu_name);
		}
	}
	
	/* all done recovering the queue */

	(void)close(fds);
	return (pq);
}
